from llm_templates.common import Conversation, Content, get_jinja_env

jinja_tpl = "{% if messages[0]['role'] == 'system' %}{% set loop_messages = messages[1:] %}{% set system_message = '<<SYS>>\n' + messages[0]['content'].strip() + '\n<</SYS>>\n\n' %}{% else %}{% set loop_messages = messages %}{% set system_message = '' %}{% endif %}{% for message in loop_messages %}{% if (message['role'] == 'user') != (loop.index0 % 2 == 0) %}{{ raise_exception('Conversation roles must alternate user/assistant/user/assistant/...') }}{% endif %}{% if loop.index0 == 0 %}{% set content = system_message + message['content'] %}{% else %}{% set content = message['content'] %}{% endif %}{% if message['role'] == 'user' %}{{ bos_token + '[INST] ' + content.strip() + ' [/INST]' }}{% elif message['role'] == 'assistant' %}{{ ' '  + content.strip() + ' ' + eos_token }}{% endif %}{% endfor %}"


def apply_template(conversation: Conversation, **kwargs):
    if kwargs.get('add_assistant_prompt', False):
        conversation.messages.append(Content(role='assistant', content=''))

    # Load template
    bos_token = "<s>"
    eos_token = "</s>"
    template = get_jinja_env().from_string(jinja_tpl)

    # Renderizar la plantilla con los mensajes proporcionados
    return template.render(messages=conversation.messages, bos_token=bos_token, eos_token=eos_token)



from pydantic import BaseModel
from typing import List, Optional
from jinja2 import Environment


class Content(BaseModel):
    role: str
    content: str


# simplified version of openai.types.chat completion_create_params.py
class Conversation(BaseModel):
    model: Optional[str]
    messages: List[Content]

    def __str__(self):
        return f"model: {self.model}, messages: {self.messages}"

    def __repr__(self):
        return f"model: {self.model}, messages: {self.messages}"

    def append_prompt(self, role: str = None, message: str = None, content: Content = None):
        if content is not None:
            self.messages.append(content)
        else:
            self.messages.append(Content(role=role, content=message))
        return self


# Function to raise exception from Jinja2 templates
def raise_exception(message):
    raise Exception(message)


def get_jinja_env(**kwargs):
    jinja_env = Environment()
    jinja_env.globals['raise_exception'] = raise_exception
    jinja_env.globals.update(kwargs)
    return jinja_env




from llm_templates.common import Conversation, get_jinja_env
import requests
import os
import json
from llm_templates.formats import zephyr, mistral, llama2, gemma

HF_URL = 'https://huggingface.co/'


class Formatter:
    def __init__(self, model_name: str = None, use_cache: bool = True,
                 clear_cache_on_create: bool = False, huggingface_api_key: str = None):
        self.model_name = model_name
        self.use_cache = use_cache
        self.huggingface_api_key = huggingface_api_key
        if clear_cache_on_create:
            clear_cache()

    def render(self, conversation: Conversation, **kwargs) -> str:
        model = self.model_name or conversation.model
        if model:
            return self._do_render(model, conversation, **kwargs)

        else:
            raise ValueError("Model name is required")

    def _get_hf_tokenizer(self, model: str) -> dict:
        tokenizer_config = None
        if self.use_cache:
            tokenizer_config = get_from_cache(model)

        if not tokenizer_config:
            # Get model tokenizer json
            headers = {}
            if self.huggingface_api_key:
                headers["Authorization"] = f"Bearer {self.huggingface_api_key}"
            url = f"{HF_URL}{model}/raw/main/tokenizer_config.json"
            response = requests.get(url, headers=headers)
            if response.status_code == 200:
                tokenizer_config = response.json()
                save_in_cache(model, tokenizer_config)
            else:
                raise ValueError(f"Model {model} configuration not found in server. "
                                 f"Some models require a token: Formatter(huggingface_api_key=HUGGINGFACE_TOKEN)")

        return tokenizer_config

    def _do_render(self, model: str, conversation: Conversation, **kwargs) -> str:
        if model not in globals():
            # Check if is in HuggingFace format: "user/model"
            is_hugging_face_fmt = model.count("/") == 1
            if is_hugging_face_fmt:
                tokenizer_config = self._get_hf_tokenizer(model)

                return hf_render(conversation=conversation, tokenizer_config=tokenizer_config, **kwargs)
            else:
                raise ValueError(f"Model {model} not found")

        template_module = globals()[model]
        return template_module.apply_template(conversation=conversation, **kwargs)


def save_in_cache(model: str, tokenizer_config: dict) -> None:
    os_home_folder = os.path.expanduser("~")
    file_name = model.replace("/", "_") + "_tokenizer_config.json"
    file_path = f"{os_home_folder}/.cache/llm_templates/{file_name}"

    if not os.path.exists(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))

    with open(file_path, "w") as f:
        f.write(json.dumps(tokenizer_config))


def get_from_cache(model: str) -> dict | None:
    os_home_folder = os.path.expanduser("~")
    file_name = model.replace("/", "_") + "_tokenizer_config.json"
    file_path = f"{os_home_folder}/.cache/llm_templates/{file_name}"

    if not os.path.exists(file_path):
        return None

    with open(file_path, "r") as f:
        return json.load(f)


def clear_cache() -> None:
    os_home_folder = os.path.expanduser("~")
    cache_folder = f"{os_home_folder}/.cache/llm_templates"
    if os.path.exists(cache_folder):
        for file in os.listdir(cache_folder):
            file_path = os.path.join(cache_folder, file)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception as e:
                print(e)


def hf_render(conversation: Conversation, tokenizer_config: dict, **kwargs) -> str:
    add_generation_prompt = kwargs.get("add_assistant_prompt", False)

    template_str = tokenizer_config.get("chat_template", '')

    template = get_jinja_env().from_string(template_str)

    for k, v in tokenizer_config.items():
        if isinstance(v, dict) and tokenizer_config[k].get('__type') == 'AddedToken':
            tokenizer_config[k] = str(tokenizer_config[k].get('content'))

    return template.render(messages=conversation.messages, **tokenizer_config,
                           add_generation_prompt=add_generation_prompt)



import setuptools
from pathlib import Path

README = (Path(__file__).parent/"README.md").read_text()

setuptools.setup(
    name="llm-templates",
    version="0.1.8",
    author="Juan Pablo Manson",
    author_email="jpmanson@gmail.com",
    description="Instruction/chat prompts creation library for text generation LLMs. It supports local and Hugging Face models.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jpmanson/llm-templates",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[],
    python_requires=">=3.9",
    license_files=("LICENSE",),
    install_requires=[
        "requests>=2.28.0",
        "Jinja2>=2.11.3",
    ]
)


from llm_templates.common import Conversation, Content, get_jinja_env


# Source format: https://ai.google.dev/gemma/docs/formatting?hl=es-419


def apply_template(conversation: Conversation, **kwargs):
    # Definir la plantilla Jinja
    template_str = "{% for message in messages %}"
    template_str += "{% if message['role'] == 'user' %}"
    template_str += "<start_of_turn>user\n{{ message['content'] }}<end_of_turn>\n"
    template_str += "{% elif message['role'] == 'assistant' %}"
    template_str += "<start_of_turn>model\n{{ message['content'] }}<end_of_turn>\n"
    template_str += "{% elif message['role'] == 'system' %}"
    template_str += "<start_of_turn>system\n{{ message['content'] }}<end_of_turn>\n"
    template_str += "{% else %}"
    template_str += "<start_of_turn>unknown\n{{ message['content'] }}<end_of_turn>\n"
    template_str += "{% endif %}"
    template_str += "{% endfor %}"
    template_str += "{% if add_generation_prompt %}"
    template_str += "<start_of_turn>model\n"
    template_str += "{% endif %}"

    template = get_jinja_env().from_string(template_str)

    # Renderizar la plantilla con los mensajes proporcionados
    return template.render(messages=conversation.messages,
                           add_generation_prompt=kwargs.get('add_assistant_prompt', False))